// This code contains NVIDIA Confidential Information and is disclosed to you
// under a form of NVIDIA software license agreement provided separately to you.
//
// Notice
// NVIDIA Corporation and its licensors retain all intellectual property and
// proprietary rights in and to this software and related documentation and
// any modifications thereto. Any use, reproduction, disclosure, or
// distribution of this software and related documentation without an express
// license agreement from NVIDIA Corporation is strictly prohibited.
//
// ALL NVIDIA DESIGN SPECIFICATIONS, CODE ARE PROVIDED "AS IS.". NVIDIA MAKES
// NO WARRANTIES, EXPRESSED, IMPLIED, STATUTORY, OR OTHERWISE WITH RESPECT TO
// THE MATERIALS, AND EXPRESSLY DISCLAIMS ALL IMPLIED WARRANTIES OF NONINFRINGEMENT,
// MERCHANTABILITY, AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Information and code furnished is believed to be accurate and reliable.
// However, NVIDIA Corporation assumes no responsibility for the consequences of use of such
// information or for any infringement of patents or other rights of third parties that may
// result from its use. No license is granted by implication or otherwise under any patent
// or patent rights of NVIDIA Corporation. Details are subject to change without notice.
// This code supersedes and replaces all information previously supplied.
// NVIDIA Corporation products are not authorized for use as critical
// components in life support devices or systems without express written approval of
// NVIDIA Corporation.
//
// Copyright (c) 2008-2013 NVIDIA Corporation. All rights reserved.
// Copyright (c) 2004-2008 AGEIA Technologies, Inc. All rights reserved.
// Copyright (c) 2001-2004 NovodeX AG. All rights reserved.  
#ifndef PX_MEMORYPOOLSTREAMS_H
#define PX_MEMORYPOOLSTREAMS_H
#include "MemoryPool.h"
#include "CmPhysXCommon.h"
#include "PxTransform.h"

#pragma warning(push)
#pragma warning (disable : 4701) // potentially uninitialized local variable used


namespace physx { namespace repx {

	template<typename TDataType>
	struct RepXDefaultValue
	{
		bool force_compile_error;
	};


#define REPX_DEFINE_DEFAULT_VALUE(type, defVal )		\
	template<>											\
	struct RepXDefaultValue<type>						\
	{													\
		type getDefaultValue() { return type(defVal); }	\
	};

	REPX_DEFINE_DEFAULT_VALUE(PxU8, 0);
	REPX_DEFINE_DEFAULT_VALUE(PxI8, 0);
	REPX_DEFINE_DEFAULT_VALUE(PxU16, 0);
	REPX_DEFINE_DEFAULT_VALUE(PxI16, 0);
	REPX_DEFINE_DEFAULT_VALUE(PxU32, 0);
	REPX_DEFINE_DEFAULT_VALUE(PxI32, 0);
	REPX_DEFINE_DEFAULT_VALUE(PxU64, 0);
	REPX_DEFINE_DEFAULT_VALUE(PxI64, 0);
	REPX_DEFINE_DEFAULT_VALUE(PxF32, 0);
	REPX_DEFINE_DEFAULT_VALUE(PxF64, 0);

#undef REPX_DEFINE_DEFAULT_VALUE

	template<>											
	struct RepXDefaultValue<PxVec3>						
	{													
		PxVec3 getDefaultValue() { return PxVec3( 0,0,0 ); }	
	};
	
	template<>											
	struct RepXDefaultValue<PxTransform>						
	{													
		PxTransform getDefaultValue() { return PxTransform::createIdentity(); }	
	};

	template<>											
	struct RepXDefaultValue<PxQuat>	
	{													
		PxQuat getDefaultValue() { return PxQuat::createIdentity(); }	
	};

/** 
 *	Mapping of PxOutputStream to a memory pool manager.
 *	Allows write-then-read semantics of a set of
 *	data.  Can safely write up to 4GB of data; then you
 *	will silently fail...
 */

template<typename TAllocatorType>
struct MemoryBufferBase : public PxOutputStream, public PxInputStream
{
	TAllocatorType* mManager;
	mutable PxU32	mWriteOffset;
	mutable PxU32	mReadOffset;
	PxU8*	mBuffer;
	PxU32	mCapacity;


	MemoryBufferBase( TAllocatorType* inManager )
		: mManager( inManager )
		, mWriteOffset( 0 )
		, mReadOffset( 0 )
		, mBuffer( NULL )
		, mCapacity( 0 )
	{
	}
	virtual						~MemoryBufferBase()
	{
		mManager->deallocate( mBuffer );
	}
	PxU8* releaseBuffer()
	{
		clear();
		mCapacity = 0;
		PxU8* retval(mBuffer);
		mBuffer = NULL;
		return retval;
	}
	void clear()
	{
		mWriteOffset = mReadOffset = 0;
	}

	virtual PxU32 read(void* dest, PxU32 count)
	{
		bool fits = ( mReadOffset + count ) <= mWriteOffset;
		PX_ASSERT( fits );
		if ( fits )
		{
			memcpy( dest, mBuffer + mReadOffset, count );
			mReadOffset += count;
			return count;
		}
		return 0;
	}

	inline void checkCapacity( PxU32 inNewCapacity )
	{
		if ( mCapacity < inNewCapacity )
		{
			PxU32 newCapacity = 32;
			while( newCapacity < inNewCapacity )
				newCapacity = newCapacity << 1;

			PxU8* newData( mManager->allocate( newCapacity ) );
			if ( mWriteOffset )
				memcpy( newData, mBuffer, mWriteOffset );
			mManager->deallocate( mBuffer );
			mBuffer = newData;
			mCapacity = newCapacity;
		}
	}

	virtual PxU32 write(const void* src, PxU32 count)
	{
		checkCapacity( mWriteOffset + count );
		memcpy( mBuffer + mWriteOffset, src, count );
		mWriteOffset += count;
		return count;
	}
};

class MemoryBuffer : public MemoryBufferBase<CMemoryPoolManager >
{
public:
	MemoryBuffer( CMemoryPoolManager* inManager ) : MemoryBufferBase<CMemoryPoolManager >( inManager ) {}
};

}}

#pragma warning(pop)
#endif